/*
  EspStrobe.cpp

  Part of WiFlash_esp32

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include <Arduino.h>
#include <AsyncUDP.h>
#include <algorithm>
#include <iterator>
#include "EspStrobe.h"

// ESP logging module
#include <esp_log.h>
static const char* TAG = "EspStrobe";

/*
   Constructor
*/
EspStrobe::EspStrobe(uint16_t address)
  : AbstractStrobe(address)
  , EspDimmer(address)
{
}

/*
   call durring setup()
*/
bool EspStrobe::begin(uint8_t pin, uint8_t pwm) {
  EspDimmer::begin(pin, pwm);

  // start update task
  String tname = "EspStrobeUpdateThread" + String(pwm);
  ESP_LOGD(TAG, "update task is named %s", tname.c_str());
  xTaskCreate(this->update,         // function name
              tname.c_str(),        // task name
              2048,                 // stack size (bytes)
              this,                 // parameter to pass
              1,                    // task priority
              NULL);                // task handle

  return true;
}

void EspStrobe::update(void * param) {
  EspStrobe *tThis = (EspStrobe *) param;
  uint32_t last_time = 0;
  TickType_t xDelay;
  for (;;) {
    xDelay = (1000 / 40.0) / portTICK_PERIOD_MS; // sleep 1 dmx frame
    // intensity changes start/stop the cycle
    if (tThis->profile_.intensity == 0) {
      ESP_LOGV(TAG, "Strobe going INACTIVE");
      tThis->state_ = STROBE_INACTIVE;
    } else {
      if (tThis->state_ == STROBE_INACTIVE) {
        ESP_LOGV(TAG, "Strobe going ACTIVE");
        tThis->state_ = STROBE_ACTIVE_ON;
      }
    }

    // time based cycle changes
    if (tThis->state_ == STROBE_ACTIVE_ON ||
        tThis->state_ == STROBE_ACTIVE_OFF) {
      uint32_t now, elapsed, period, dur;
      now = micros();
      elapsed = now - last_time;
      period = tThis->dmxPeriod(tThis->profile_.rate);
      dur = tThis->dmxDuration(tThis->profile_.duration);

      if (elapsed > period ||         // hz period completed
          now < last_time) {      // micros() wraps 32 bits every ~70 min.
        // cycle restarts
        tThis->state_ = STROBE_ACTIVE_ON;
        last_time = now;
        // yield task until end of duration
        xDelay = std::min(xDelay, dur / portTICK_PERIOD_MS);
      } else if (elapsed > dur) {    // durration period completed
        // cycle enters dark phase
        tThis->state_ = STROBE_ACTIVE_OFF;
        // yield task until end of period
        xDelay = std::min(xDelay, (period - dur) / portTICK_PERIOD_MS);
      } else {
        if (tThis->state_ == STROBE_ACTIVE_ON) {
          // yield task for remainder of duration
          xDelay = std::min(xDelay, (dur - elapsed) / portTICK_PERIOD_MS);
        } else {
          // yield task for remainder of period
          xDelay = std::min(xDelay, (period - elapsed) / portTICK_PERIOD_MS);
        }
      }
    }

    // set LED output
    if (tThis->state_ == STROBE_ACTIVE_ON) {
      tThis->setLevel(tThis->profile_.intensity);
    } else {
      tThis->setLevel(0);
    }

    // yield task
    vTaskDelay(xDelay);
  }
  vTaskDelete( NULL );
}
